"""This module holds utils to be used by transformations."""

from butterfree.transform.utils.function import Function
from butterfree.transform.utils.window_spec import Window

__all__ = ["Window", "Function"]



import datetime
import importlib
import inspect
import logging
import os
import pkgutil
import sys
from typing import Set

import boto3
import setuptools
import typer
from botocore.exceptions import ClientError

from butterfree.configs import environment
from butterfree.migrations.database_migration import ALLOWED_DATABASE
from butterfree.pipelines import FeatureSetPipeline

app = typer.Typer(
    help="Apply the automatic migrations in a database.", no_args_is_help=True
)

logger = logging.getLogger(__name__)


def __find_modules(path: str) -> Set[str]:
    modules = set()
    for pkg in setuptools.find_packages(path):
        modules.add(pkg)
        pkg_path = path + "/" + pkg.replace(".", "/")

        # different usage for older python3 versions
        if sys.version_info.minor < 6:
            for _, name, is_pkg in pkgutil.iter_modules([pkg_path]):
                if not is_pkg:
                    modules.add(pkg + "." + name)
        else:
            for info in pkgutil.iter_modules([pkg_path]):
                if not info.ispkg:
                    modules.add(pkg + "." + info.name)
    return modules


def __fs_objects(path: str) -> Set[FeatureSetPipeline]:
    logger.info(f"Looking for python modules under {path}...")
    modules = __find_modules(path)
    if not modules:
        logger.error(f"Path: {path} not found!")
        return set()

    logger.info("Importing modules...")
    package = ".".join(path.strip("/").split("/"))
    imported = set(
        importlib.import_module(f".{name}", package=package) for name in modules
    )

    logger.info("Scanning modules...")
    content = {
        module: set(
            filter(
                lambda x: not x.startswith("__"),  # filter "__any__" attributes
                set(item for item in dir(module)),
            )
        )
        for module in imported
    }

    instances = set()
    for module, items in content.items():
        for item in items:
            value = getattr(module, item)
            if not value:
                continue

            # filtering non-classes
            if not inspect.isclass(value):
                continue

            # filtering abstractions
            if inspect.isabstract(value):
                continue

            # filtering classes that doesn't inherit from FeatureSetPipeline
            if not issubclass(value, FeatureSetPipeline):
                continue

            # filtering FeatureSetPipeline itself
            if value == FeatureSetPipeline:
                continue

            instances.add(value)

    logger.info("Creating instances...")
    return set(value() for value in instances)  # type: ignore


PATH = typer.Argument(
    ...,
    help="Full or relative path to where feature set pipelines are being defined.",
)

GENERATE_LOGS = typer.Option(
    False, help="To generate the logs in local file 'logging.json'."
)

DEBUG_MODE = typer.Option(
    False,
    help="To view the queries resulting from the migration, DON'T apply the migration.",
)


class Migrate:
    """Execute migration operations in a Database based on pipeline Writer.

    Attributes:
        pipelines: list of Feature Set Pipelines to use to migration.
    """

    def __init__(
        self,
        pipelines: Set[FeatureSetPipeline],
    ) -> None:
        self.pipelines = pipelines

    def _send_logs_to_s3(self, file_local: bool, debug_mode: bool) -> None:
        """Send all migration logs to S3."""
        file_name = "../logging.json"

        if not file_local and os.path.exists(file_name):
            s3_client = boto3.client("s3")

            timestamp = datetime.datetime.now()

            if debug_mode:
                object_name = (
                    f"logs/migrate-debug-mode/"
                    f"{timestamp.strftime('%Y-%m-%d')}"
                    f"/logging-{timestamp.strftime('%H:%M:%S')}.json"
                )
            else:
                object_name = (
                    f"logs/migrate/"
                    f"{timestamp.strftime('%Y-%m-%d')}"
                    f"/logging-{timestamp.strftime('%H:%M:%S')}.json"
                )
            bucket = environment.get_variable("FEATURE_STORE_S3_BUCKET")

            try:
                s3_client.upload_file(
                    file_name,
                    bucket,
                    object_name,
                    ExtraArgs={"ACL": "bucket-owner-full-control"},
                )
            except ClientError:
                raise

            os.remove(file_name)
        elif os.path.exists(file_name):
            print("Logs written to ../logging.json")
        else:
            print("No logs were generated.")

    def run(self, generate_logs: bool = False, debug_mode: bool = False) -> None:
        """Construct and apply the migrations."""
        for pipeline in self.pipelines:
            for writer in pipeline.sink.writers:
                db = writer.db_config.database
                if db == "cassandra":
                    migration = ALLOWED_DATABASE[db]
                    migration.apply_migration(pipeline.feature_set, writer, debug_mode)
                else:
                    logger.warning(f"Butterfree not supporting {db} Migrations yet.")

        self._send_logs_to_s3(generate_logs, debug_mode)


@app.command("apply")
def migrate(
    path: str = PATH, generate_logs: bool = GENERATE_LOGS, debug_mode: bool = DEBUG_MODE
) -> Set[FeatureSetPipeline]:
    """Scan and run database migrations for feature set pipelines defined under PATH.

    Butterfree will scan a given path for classes that inherit from its
    FeatureSetPipeline and create dry instances of it to extract schema and writer
    information. By doing this, Butterfree can compare all defined feature set schemas
    to their current state on each sink being used.

    All pipelines must be under python modules inside path, so we can dynamically
    import and instantiate them.
    """
    pipe_set = __fs_objects(path)
    Migrate(pipe_set).run(generate_logs, debug_mode)
    return pipe_set



"""Replace transformer for dataframes."""

from itertools import chain
from typing import Dict

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import coalesce, col, create_map, lit


def replace(
    dataframe: DataFrame, column: str, replace_dict: Dict[str, str]
) -> DataFrame:
    """Replace values of a string column in the dataframe using a dict.

    Example:

    >>> from butterfree.extract.pre_processing import replace
    ... from butterfree.testing.dataframe import (
    ...     assert_dataframe_equality,
    ...     create_df_from_collection,
    ... )
    >>> from pyspark import SparkContext
    >>> from pyspark.sql import session
    >>> spark_context = SparkContext.getOrCreate()
    >>> spark_session = session.SparkSession(spark_context)
    >>> input_data = [
    ...     {"id":1, "type": "a"}, {"id":2, "type": "b"}, {"id":3, "type": "c"}
    ... ]
    >>> input_df = create_df_from_collection(input_data, spark_context, spark_session)
    >>> input_df.collect()

    [Row(id=1, type='a'), Row(id=2, type='b'), Row(id=3, type='c')]

    >>> replace_dict = {"a": "type_a", "b": "type_b"}
    >>> replace(input_df, "type", replace_dict).collect()

    [Row(id=1, type='type_a'), Row(id=2, type='type_b'), Row(id=3, type='c')]

    Args:
        dataframe: data to be transformed.
        column: string column on the dataframe where to apply the replace.
        replace_dict: dict with values to be replaced.
            All mapped values must be string.

    Returns:
        Dataframe with column values replaced.

    """
    if not isinstance(dataframe, DataFrame):
        raise ValueError("dataframe needs to be a Pyspark DataFrame type")
    if (column not in dict(dataframe.dtypes)) or (
        dict(dataframe.dtypes)[column] != "string"
    ):
        raise ValueError("column needs to be the name of an string column in dataframe")
    if (not isinstance(replace_dict, dict)) or (
        not all(isinstance(value, str) for value in chain(*replace_dict.items()))
    ):
        raise ValueError(
            "replace_dict needs to be a Python dict with "
            "all keys and values as string values"
        )

    mapping = create_map(
        [lit(value) for value in chain(*replace_dict.items())]  # type: ignore
    )
    return dataframe.withColumn(column, coalesce(mapping[col(column)], col(column)))



"""Holds data loaders for historical and online feature store."""

from butterfree.load.writers.delta_writer import DeltaWriter
from butterfree.load.writers.historical_feature_store_writer import (
    HistoricalFeatureStoreWriter,
)
from butterfree.load.writers.online_feature_store_writer import OnlineFeatureStoreWriter

__all__ = ["HistoricalFeatureStoreWriter", "OnlineFeatureStoreWriter", "DeltaWriter"]



"""Holds the SourceSelector class."""

from typing import List, Optional

from pyspark.sql import DataFrame

from butterfree.clients import SparkClient
from butterfree.extract.readers.reader import Reader
from butterfree.hooks import HookableComponent


class Source(HookableComponent):
    """The definition of the the entry point data for the ETL pipeline.

    A FeatureSet (the next step in the pipeline) expects a single dataframe as
    input. This dataframe is built from a data composition of one or more
    readers defined in the Source. There is only one Source for pipeline.

    TODO refactor query into multiple query components
    TODO make it harder to do query injection

    Attributes:
        readers: list of readers from where the source will get data.
        query: Spark SQL query to run against the readers.

    Example:
        Simple example regarding Source class instantiation.

    >>> from butterfree.extract import Source
    >>> from butterfree.extract.readers import TableReader, FileReader
    >>> from butterfree.clients import SparkClient
    >>> spark_client = SparkClient()
    >>> source = Source(
    ...    readers=[
    ...        TableReader(
    ...            id="table_reader_id",
    ...            database="table_reader_db",
    ...            table="table_reader_table",
    ...        ),
    ...        FileReader(id="file_reader_id", path="data_sample_path", format="json"),
    ...    ],
    ...    query=f"select a.*, b.feature2 "
    ...    f"from table_reader_id a "
    ...    f"inner join file_reader_id b on a.id = b.id ",
    ...)
    >>> df = source.construct(spark_client)

        This last method will use the Spark Client, as default, to create
        temporary views regarding each reader and, after, will run the
        desired query and return a dataframe.

        The `eager_evaluation` param forces Spark to apply the currently
        mapped changes to the DataFrame. When this parameter is set to
        False, Spark follows its standard behaviour of lazy evaluation.
        Lazy evaluation can improve Spark's performance as it allows
        Spark to build the best version of the execution plan.

    """

    def __init__(
        self,
        readers: List[Reader],
        query: str,
        eager_evaluation: bool = True,
    ) -> None:
        super().__init__()
        self.enable_pre_hooks = False
        self.readers = readers
        self.query = query
        self.eager_evaluation = eager_evaluation

    def construct(
        self,
        client: SparkClient,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
    ) -> DataFrame:
        """Construct an entry point dataframe for a feature set.

        This method will assemble multiple readers, by building each one and
        querying them using a Spark SQL. It's important to highlight that in
        order to filter a dataframe regarding date boundaries, it's important
        to define a IncrementalStrategy, otherwise your data will not be filtered.
        Besides, both start and end dates parameters are optional.

        After that, there's the caching of the dataframe, however since cache()
        in Spark is lazy, an action is triggered in order to force persistence.

        Args:
            client: client responsible for connecting to Spark session.
            start_date: user defined start date for filtering.
            end_date: user defined end date for filtering.

        Returns:
            DataFrame with the query result against all readers.

        """
        for reader in self.readers:
            reader.build(
                client=client, start_date=start_date, end_date=end_date
            )  # create temporary views for each reader

        dataframe = client.sql(self.query)

        if not dataframe.isStreaming and self.eager_evaluation:
            dataframe.cache().count()

        post_hook_df = self.run_post_hooks(dataframe)

        return post_hook_df



"""Module defining partitioning methods."""

from typing import Any, Dict, List

from pyspark.sql import DataFrame


def extract_partition_values(
    dataframe: DataFrame, partition_columns: List[str]
) -> List[Dict[str, Any]]:
    """Extract distinct partition values from a given dataframe.

    Args:
        dataframe: dataframe from where to extract partition values.
        partition_columns: name of partition columns presented on the dataframe.

    Returns:
        distinct partition values.
    """
    return (
        dataframe.select(*partition_columns)
        .distinct()
        .rdd.map(lambda row: row.asDict(True))
        .collect()
    )



"""Holds available database migrations."""

from butterfree.migrations.database_migration.cassandra_migration import (
    CassandraMigration,
)
from butterfree.migrations.database_migration.database_migration import Diff
from butterfree.migrations.database_migration.metastore_migration import (
    MetastoreMigration,
)

__all__ = ["CassandraMigration", "MetastoreMigration", "Diff"]


ALLOWED_DATABASE = {
    "cassandra": CassandraMigration(),
    "metastore": MetastoreMigration(),
}



"""Abstract class for database clients."""

from abc import ABC, abstractmethod
from typing import Any, Optional


class AbstractClient(ABC):
    """Abstract base class for database clients."""

    @property
    @abstractmethod
    def conn(self) -> Any:
        """Returns a connection object."""
        pass

    @abstractmethod
    def sql(self, query: str) -> Any:
        """Runs a query.

        Args:
          query: client query.

        Returns:
            Set of records.
        """
        pass

    @abstractmethod
    def get_schema(self, table: str, database: Optional[str] = None) -> Any:
        """Returns desired table schema.

        Attributes:
            table: desired table.

        Returns:
            A list of dictionaries in the format
            [{"column_name": "example1", type: "Spark_type"}, ...]

        """
        pass



"""Stack Transform entity."""

import re
from typing import List

from pyspark.sql import DataFrame
from pyspark.sql.functions import array, explode

from butterfree.transform.transformations.transform_component import TransformComponent


class StackTransform(TransformComponent):
    """Defines a Stack transformation.

    For instantiation it is needed the name of the columns or a pattern to use
    to find the columns that need to be stacked. This transform generates just
    one column as output.

    Attributes:
        columns_names: full names or patterns to search for target columns on
            the dataframe.
            By default a single `*` character is considered a wildcard and can
            be anywhere in the string, multiple wildcards are not supported.
            Strings can also start with an `!` (exclamation mark), it indicates
            a negation, be it a regular string or simple pattern.
            When parameter :param is_regex: is `True`, simple patterns wildcards
            and negation are disabled and all strings are interpreted as regular
            expressions.

        is_regex: boolean flag to indicate if columns_names passed are a Python
            regex string patterns.

    Example:
        >>> from pyspark import SparkContext
        >>> from pyspark.sql import session
        >>> from butterfree.testing.dataframe import create_df_from_collection
        >>> from butterfree.constants import DataType
        >>> from butterfree.transform.transformations import StackTransform
        >>> from butterfree.transform.features import Feature
        >>> spark_context = SparkContext.getOrCreate()
        >>> spark_session = session.SparkSession(spark_context)
        >>> data = [
        ...    {"feature": 100, "id_a": 1, "id_b": 2},
        ...    {"feature": 120, "id_a": 3, "id_b": 4},
        ... ]
        >>> df = create_df_from_collection(data, spark_context, spark_session)
        >>> df.collect()
        [Row(feature=100, id_a=1, id_b=2), Row(feature=120, id_a=3, id_b=4)]
        >>> feature = Feature(
        ...     name="stack_ids",
        ...     description="id_a and id_b stacked in a single column.",
        ...     dtype=DataType.BIGINT,
        ...     transformation=StackTransform("id_a", "id_b"),
        ... )
        >>> feature.transform(df).collect()
        [
            Row(feature=100, id_a=1, id_b=2, stack_ids=1),
            Row(feature=100, id_a=1, id_b=2, stack_ids=2),
            Row(feature=120, id_a=3, id_b=4, stack_ids=3),
            Row(feature=120, id_a=3, id_b=4, stack_ids=4)
        ]

        The StackTransform can be instantiated using a column pattern instead of
        the columns full names. Like this way:

        >>> feature = Feature(
        ...     name="stack_ids",
        ...     description="id_a and id_b stacked in a single column.",
        ...     dtype=DataType.BIGINT,
        ...     transformation=StackTransform(columns_prefix="id_*"),
        ... )

    """

    def __init__(self, *columns_names: str, is_regex: bool = False):
        super().__init__()
        self.columns_names = columns_names
        self.regex_enabled = is_regex

    @property
    def output_columns(self) -> List[str]:
        """Columns generated by the transformation."""
        return [self._parent.name]

    def _matches_pattern(self, pattern: str, column: str) -> bool:
        """Verify if the column name matches the pattern.

        Args:
            pattern: string pattern to use.
            column: column names to try match with the pattern.

        Returns:
            True for a column that matches the pattern, False otherwise.

        """
        if self.regex_enabled:
            return bool(re.match(pattern, column))
        negate = False
        if pattern.startswith("!"):
            negate = True
            pattern = pattern[1:]
        split = pattern.split("*")
        matches_pattern = column.startswith(split[0]) and column.endswith(split[-1])
        return matches_pattern if not negate else not matches_pattern

    def transform(self, dataframe: DataFrame) -> DataFrame:
        """Performs a transformation to the feature pipeline.

        Args:
            dataframe: input dataframe.

        Returns:
            Transformed dataframe.

        """
        columns = [
            column
            for column in dataframe.columns
            if any(
                self._matches_pattern(pattern, column) for pattern in self.columns_names
            )
        ]

        if not columns:
            raise ValueError(
                "Columns not found, columns in df: {}".format(dataframe.columns)
            )

        return dataframe.withColumn(self._parent.name, explode(array(*columns)))



"""Validation implementing basic checks over the dataframe."""

from typing import Optional

from pyspark.sql.dataframe import DataFrame

from butterfree.constants.columns import TIMESTAMP_COLUMN
from butterfree.validations.validation import Validation


class BasicValidation(Validation):
    """Basic validation suite for Feature Set's dataframe.

    Attributes:
        dataframe: object to be verified

    """

    def __init__(self, dataframe: Optional[DataFrame] = None):
        super().__init__(dataframe)

    def check(self) -> None:
        """Check basic validation properties about the dataframe.

        Raises:
            ValueError: if any of the verifications fail

        """
        self.validate_df_is_spark_df()
        self.validate_column_ts()
        self.validate_df_is_empty()

    def validate_column_ts(self) -> None:
        """Check dataframe's ts column.

        Raises:
            ValueError: if dataframe don't have a column named ts.

        """
        if not self.dataframe:
            raise ValueError("DataFrame can't be None.")
        if TIMESTAMP_COLUMN not in self.dataframe.columns:
            raise ValueError(f"DataFrame must have a '{TIMESTAMP_COLUMN}' column.")

    def validate_df_is_empty(self) -> None:
        """Check dataframe emptiness.

        Raises:
            ValueError: if dataframe is empty and is not streaming.

        """
        if not self.dataframe:
            raise ValueError("DataFrame can't be None.")
        if (not self.dataframe.isStreaming) and self.dataframe.rdd.isEmpty():
            raise ValueError("DataFrame can't be empty.")

    def validate_df_is_spark_df(self) -> None:
        """Check type of dataframe object.

        Raises:
            ValueError: if dataframe is not instance of pyspark.sql.DataFrame.

        """
        if not self.dataframe:
            raise ValueError("DataFrame can't be None.")
        if not isinstance(self.dataframe, DataFrame):
            raise ValueError(
                "dataframe needs to be a instance of pyspark.sql.DataFrame"
            )



import re
from dataclasses import dataclass
from typing import List, Optional, Tuple

from pyspark.sql import DataFrame

from butterfree.constants.data_type import DataType

BUTTERFREE_DTYPES = {
    "string": DataType.STRING.spark_sql,
    "long": DataType.BIGINT.spark_sql,
    "double": DataType.DOUBLE.spark_sql,
    "boolean": DataType.BOOLEAN.spark_sql,
    "integer": DataType.INTEGER.spark_sql,
    "date": DataType.DATE.spark_sql,
    "timestamp": DataType.TIMESTAMP.spark_sql,
    "array": {
        "long": DataType.ARRAY_BIGINT.spark_sql,
        "float": DataType.ARRAY_FLOAT.spark_sql,
        "string": DataType.ARRAY_STRING.spark_sql,
    },
}


@dataclass(frozen=True)
class Table:  # noqa: D101
    id: str
    database: str
    name: str


class FeatureSetCreation:
    """Class to auto-generate readers and features."""

    def _get_features_with_regex(self, sql_query: str) -> List[str]:
        features = []
        sql_query = " ".join(sql_query.split())
        first_pattern = re.compile("[(]?([\w.*]+)[)]?,", re.IGNORECASE)
        second_pattern = re.compile("(\w+)\s(from)", re.IGNORECASE)

        for pattern in [first_pattern, second_pattern]:
            matches = pattern.finditer(sql_query)
            for match in matches:
                feature = match.group(1)

                if "." in feature:
                    feature = feature.split(".")[1]

                features.append(feature)

        return features

    def _get_data_type(self, field_name: str, df: DataFrame) -> str:
        for field in df.schema.jsonValue()["fields"]:
            if field["name"] == field_name:

                field_type = field["type"]

                if isinstance(field_type, dict):

                    field_type_keys = field_type.keys()

                    if "type" in field_type_keys and "elementType" in field_type_keys:
                        return (
                            "."
                            + BUTTERFREE_DTYPES[field_type["type"]][  # type: ignore
                                field_type["elementType"]
                            ]
                        )

                return "." + BUTTERFREE_DTYPES[field["type"]]

        return ""

    def _get_tables_with_regex(self, sql_query: str) -> Tuple[List[Table], str]:

        modified_sql_query = sql_query
        tables = []
        stop_words = [
            "left",
            "right",
            "full outer",
            "inner",
            "where",
            "join",
            "on",
            "as",
        ]
        keywords = ["from", "join"]

        for keyword in keywords:
            pattern = re.compile(
                rf"\b{keyword}\s+(\w+\.\w+|\w+)\s+(\w+)", re.IGNORECASE
            )
            matches = pattern.finditer(sql_query)

            for match in matches:
                full_table_name = match.group(1)
                id = match.group(2).strip()

                if id in stop_words:
                    id = full_table_name

                if "." in full_table_name:
                    database, table = full_table_name.split(".")

                    modified_sql_query = re.sub(
                        rf"\b{database}\.{table}\b", table, modified_sql_query
                    )

                    tables.append(Table(id=id, database=database, name=table))
                else:
                    modified_sql_query = re.sub(
                        rf"\b{full_table_name}\b", full_table_name, modified_sql_query
                    )
                    tables.append(Table(id=id, database="TBD", name=full_table_name))

        return tables, modified_sql_query

    def get_readers(self, sql_query: str) -> str:
        """
        Extracts table readers from a SQL query and formats them as a string.

        Args:
            sql_query (str): The SQL query from which to extract table readers.

        Returns:
            str: A formatted string containing the table readers.
        """
        tables, modified_sql_query = self._get_tables_with_regex(sql_query.lower())
        readers = []
        for table in tables:
            table_reader_string = f"""
            TableReader(
                id="{table.id}",
                database="{table.database}",
                table="{table.name}"
            ),
            """
            readers.append(table_reader_string)

        final_string = """
        source=Source(
            readers=[
            {}
            ],
            query=(
            \"\"\"
            {}
            \"\"\"
            ),
        ),
        """.format(
            "".join(readers), modified_sql_query.replace("\n", "\n\t\t")
        )

        return final_string

    def get_features(self, sql_query: str, df: Optional[DataFrame] = None) -> str:
        """
        Extract features from a SQL query and return them formatted as a string.

        Args:
            sql_query (str): The SQL query used to extract features.
            df (Optional[DataFrame], optional): Optional DataFrame used to infer data types. Defaults to None.

        Returns:
            str: A formatted string containing the extracted features.

        This sould be used on Databricks.

        Especially if you want automatic type inference without passing a reference dataframe.
        The utility will only work in an environment where a spark session is available in the environment
        """  # noqa: E501

        features = self._get_features_with_regex(sql_query)
        features_formatted = []
        for feature in features:
            description = feature.replace("__", " ").replace("_", " ").capitalize()

            data_type = "."

            if df is None:
                df = spark.sql(sql_query)  # type: ignore # noqa: F821

            data_type = self._get_data_type(feature, df)

            feature_string = f"""
            Feature(
            name="{feature}",
            description="{description}",
            dtype=DataType{data_type},
            ),
            """
            features_formatted.append(feature_string)

        final_string = ("features=[\t{}],\n),").format("".join(features_formatted))

        return final_string



"""Holds Schema Compatibility Hooks definitions."""

from butterfree.hooks.schema_compatibility.cassandra_table_schema_compatibility_hook import (  # noqa
    CassandraTableSchemaCompatibilityHook,
)
from butterfree.hooks.schema_compatibility.spark_table_schema_compatibility_hook import (  # noqa
    SparkTableSchemaCompatibilityHook,
)

__all__ = ["SparkTableSchemaCompatibilityHook", "CassandraTableSchemaCompatibilityHook"]



import logging

from delta.tables import DeltaTable
from pyspark.sql.dataframe import DataFrame

from butterfree.clients import SparkClient

logger = logging.getLogger(__name__)


class DeltaWriter:
    """Control operations on Delta Tables.

    Resposible for merging and optimizing.
    """

    @staticmethod
    def _get_full_table_name(table, database):
        if database:
            return "{}.{}".format(database, table)
        else:
            return table

    @staticmethod
    def _convert_to_delta(client: SparkClient, table: str):
        logger.info(f"Converting {table} to Delta...")
        client.conn.sql(f"CONVERT TO DELTA {table}")
        logger.info("Conversion done.")

    @staticmethod
    def merge(
        client: SparkClient,
        database: str,
        table: str,
        merge_on: list,
        source_df: DataFrame,
        when_not_matched_insert_condition: str = None,
        when_matched_update_condition: str = None,
        when_matched_delete_condition: str = None,
    ):
        """
        Merge a source dataframe to a Delta table.

        By default, it will update when matched, and insert when
        not matched (simple upsert).

        You can change this behavior by setting:
        - when_not_matched_insert_condition: it will only insert
            when this specified condition is true
        - when_matched_update_condition: it will only update when this
            specified condition is true. You can refer to the columns
        in the source dataframe as source.<column_name>, and the columns
            in the target table as target.<column_name>.
        - when_matched_delete_condition: it will add an operation to delete,
            but only if this condition is true. Again, source and
            target dataframe columns can be referred to respectively as
            source.<column_name> and target.<column_name>
        """
        try:
            full_table_name = DeltaWriter._get_full_table_name(table, database)

            table_exists = client.conn.catalog.tableExists(full_table_name)

            if table_exists:
                pd_df = client.conn.sql(
                    f"DESCRIBE TABLE EXTENDED {full_table_name}"
                ).toPandas()
                provider = (
                    pd_df.reset_index()
                    .groupby(["col_name"])["data_type"]
                    .aggregate("first")
                    .Provider
                )
                table_is_delta = provider.lower() == "delta"

                if not table_is_delta:
                    DeltaWriter()._convert_to_delta(client, full_table_name)

            # For schema evolution
            client.conn.conf.set(
                "spark.databricks.delta.schema.autoMerge.enabled", "true"
            )

            target_table = DeltaTable.forName(client.conn, full_table_name)
            join_condition = " AND ".join(
                [f"source.{col} = target.{col}" for col in merge_on]
            )
            merge_builder = target_table.alias("target").merge(
                source_df.alias("source"), join_condition
            )
            if when_matched_delete_condition:
                merge_builder = merge_builder.whenMatchedDelete(
                    condition=when_matched_delete_condition
                )

            merge_builder.whenMatchedUpdateAll(
                condition=when_matched_update_condition
            ).whenNotMatchedInsertAll(
                condition=when_not_matched_insert_condition
            ).execute()
        except Exception as e:
            logger.error(f"Merge operation on {full_table_name} failed: {e}")

    @staticmethod
    def vacuum(table: str, retention_hours: int, client: SparkClient):
        """Vacuum a Delta table.

        Vacuum remove unused files (files not managed by Delta + files
        that are not in the latest state).
        After vacuum it's impossible to time travel to versions
        older than the `retention` time.
        Default retention is 7 days. Lower retentions will be warned,
        unless it's set to false.
        Set spark.databricks.delta.retentionDurationCheck.enabled
        to false for low retentions.
        https://docs.databricks.com/en/sql/language-manual/delta-vacuum.html
        """

        command = f"VACUUM {table} RETAIN {retention_hours} HOURS"
        logger.info(f"Running vacuum with command {command}")
        client.conn.sql(command)
        logger.info(f"Vacuum successful for table {table}")

    @staticmethod
    def optimize(
        client: SparkClient,
        table: str = None,
        z_order: list = None,
        date_column: str = "timestamp",
        from_date: str = None,
        auto_compact: bool = False,
        optimize_write: bool = False,
    ):
        """Optimize a Delta table.

        For auto-compaction and optimize write DBR >= 14.3 LTS
        and Delta >= 3.1.0 are MANDATORY.
        For z-ordering DBR >= 13.3 LTS and Delta >= 2.0.0 are MANDATORY.
        Auto-compaction (recommended) reduces the small file problem
        (overhead due to lots of metadata).
        Z-order by columns that is commonly used in queries
        predicates and has a high cardinality.
        https://docs.delta.io/latest/optimizations-oss.html
        """

        if auto_compact:
            client.conf.set("spark.databricks.delta.autoCompact.enabled", "true")

        if optimize_write:
            client.conf.set("spark.databricks.delta.optimizeWrite.enabled", "true")

        if table:
            command = f"OPTIMIZE {table}"

            if from_date:
                command += f"WHERE {date_column} >= {from_date}"

            if z_order:
                command += f" ZORDER BY {','.join(z_order)}"

            logger.info(f"Running optimize with command {command}...")
            client.conn.sql(command)
            logger.info(f"Optimize successful for table {table}.")



"""CustomTransform entity."""

from typing import Any, Callable, List

from pyspark.sql import DataFrame

from butterfree.transform.transformations.transform_component import TransformComponent


class CustomTransform(TransformComponent):
    """Defines a Custom Transform.

    Attributes:
        transformer: function to use for transforming the dataframe
        **kwargs: kwargs for the transformer

    Example:
        It's necessary to instantiate the CustomTransform class using
        a custom method that must always receive a dataframe and the
        parent feature as arguments and the custom arguments must be
        passed to the builder through *kwargs.

        >>> from butterfree.transform.transformations import CustomTransform
        >>> from butterfree.transform.features import Feature
        >>> from butterfree.constants import DataType
        >>> from pyspark import SparkContext
        >>> from pyspark.sql import session
        >>> import pyspark.sql.functions as F
        >>> sc = SparkContext.getOrCreate()
        >>> spark = session.SparkSession(sc)
        >>> df = spark.createDataFrame([(1, "2016-04-11 11:31:11", 200, 200),
        ...                             (1, "2016-04-11 11:44:12", 300, 300),
        ...                             (1, "2016-04-11 11:46:24", 400, 400),
        ...                             (1, "2016-04-11 12:03:21", 500, 500)]
        ...                           ).toDF("id", "timestamp", "feature1", "feature2")
        >>> def divide(df, parent_feature, column1, column2):
        ...     name = parent_feature.get_output_columns()[0]
        ...     df = df.withColumn(name, F.col(column1) / F.col(column2))
        ...     return df
        >>> feature = Feature(
        ...    name="feature",
        ...    description="custom transform usage example",
        ...    dtype=DataType.DOUBLE,
        ...    transformation=CustomTransform(
        ...        transformer=divide, column1="feature1", column2="feature2",
        ...    )
        ...)
        >>> feature.transform(df).orderBy("timestamp").show()
        +--------+--------+---+-------------------+--------+
        |feature1|feature2| id|          timestamp|feature|
        +--------+--------+---+-------------------+--------+
        |     200|     200|  1|2016-04-11 11:31:11|    1.0|
        |     300|     300|  1|2016-04-11 11:44:12|    1.0|
        |     400|     400|  1|2016-04-11 11:46:24|    1.0|
        |     500|     500|  1|2016-04-11 12:03:21|    1.0|
        +--------+--------+---+-------------------+--------+

    """

    def __init__(self, transformer: Callable[..., Any], **kwargs: Any):
        super().__init__()
        self.transformer = transformer
        self.transformer__kwargs = kwargs

    @property
    def transformer(self) -> Callable[..., Any]:
        """Function to use for transforming the dataframe."""
        return self._transformer

    @transformer.setter
    def transformer(self, method: Callable[..., Any]) -> None:
        if method is None:
            raise ValueError("A method must be provided to CustomTransform")
        self._transformer = method

    @property
    def output_columns(self) -> List[str]:
        """Columns generated by transformation."""
        return [self._parent.name]

    def transform(self, dataframe: DataFrame) -> DataFrame:
        """Performs a transformation to the feature pipeline.

        Args:
            dataframe: input dataframe to be transformed.

        Returns:
            Transformed dataframe.

        """
        dataframe = self.transformer(
            dataframe,
            self.parent,
            **self.transformer__kwargs,
        )
        return dataframe



"""FeatureSetPipeline entity."""

from typing import List, Optional

from butterfree.clients import SparkClient
from butterfree.dataframe_service import repartition_sort_df
from butterfree.extract import Source
from butterfree.load import Sink
from butterfree.transform import FeatureSet


class FeatureSetPipeline:
    """Defines a ETL pipeline for the construction of a feature set.

    Attributes:
        source: source of the data, the entry point of the pipeline.
        feature_set: feature set composed by features and context metadata.
        sink: sink used to write the output dataframe in the desired locations.
        spark_client: client used to access Spark connection.

    Example:
        This an example regarding the feature set pipeline definition. All
        sources, feature set (and its features) and writers are defined.

        >>> import os

        >>> from butterfree.pipelines import FeatureSetPipeline
        >>> from butterfree.constants.columns import TIMESTAMP_COLUMN
        >>> from butterfree.configs.db import MetastoreConfig
        >>> from butterfree.extract import Source
        >>> from butterfree.extract.readers import TableReader
        >>> from butterfree.transform import FeatureSet
        >>> from butterfree.transform.features import (
        ...     Feature,
        ...     KeyFeature,
        ...     TimestampFeature,
        ...)
        >>> from butterfree.transform.transformations import (
        ...     SparkFunctionTransform,
        ...     CustomTransform,
        ... )
        >>> from butterfree.load import Sink
        >>> from butterfree.load.writers import HistoricalFeatureStoreWriter
        >>> from pyspark.sql import functions

        >>> def divide(df, fs, column1, column2):
        ...     name = fs.get_output_columns()[0]
        ...     df = df.withColumn(name,
        ...            functions.col(column1) / functions.col(column2))
        ...     return df

        >>> pipeline = FeatureSetPipeline(
        ...    source=Source(
        ...        readers=[
        ...            TableReader(
        ...                id="table_reader_id",
        ...                database="table_reader_db",
        ...                table="table_reader_table",
        ...            ),
        ...        ],
        ...        query=f"select * from table_reader_id ",
        ...    ),
        ...    feature_set=FeatureSet(
        ...        name="feature_set",
        ...        entity="entity",
        ...        description="description",
        ...        features=[
        ...            Feature(
        ...                name="feature1",
        ...                description="test",
        ...            transformation=SparkFunctionTransform(
        ...                 functions=[Function(functions.avg, DataType.DOUBLE),
        ...                         Function(functions.stddev_pop, DataType.DOUBLE)],
        ...             ).with_window(
        ...                 partition_by="id",
        ...                 order_by=TIMESTAMP_COLUMN,
        ...                 mode="fixed_windows",
        ...                 window_definition=["2 minutes", "15 minutes"],
        ...             ),
        ...            ),
        ...            Feature(
        ...                name="divided_feature",
        ...                description="unit test",
        ...                transformation=CustomTransform(
        ...                    transformer=divide,
        ...                    column1="feature1",
        ...                    column2="feature2",
        ...                ),
        ...            ),
        ...        ],
        ...        keys=[
        ...            KeyFeature(
        ...                name="id",
        ...                description="The user's Main ID or device ID"
        ...            )
        ...        ],
        ...        timestamp=TimestampFeature(),
        ...    ),
        ...    sink=Sink(
        ...         writers=[
        ...            HistoricalFeatureStoreWriter(
        ...                db_config=MetastoreConfig(
        ...                    format_="parquet",
        ...                    path=os.path.join(
        ...                        os.path.dirname(os.path.abspath(__file__))
        ...                    ),
        ...                ),
        ...            )
        ...        ],
        ...    ),
        ...)

        >>> pipeline.run()

        This last method (run) will execute the pipeline flow, it'll read from
        the defined sources, compute all the transformations and save the data
        to the specified locations.

         We can run the pipeline over a range of dates by passing an end-date
        and a start-date, where it will only bring data within this date range.

        >>> pipeline.run(end_date="2020-08-04", start_date="2020-07-04")

        Or run up to a date, where it will only bring data up to the specific date.

        >>> pipeline.run(end_date="2020-08-04")

        Or just a specific date, where you will only bring data for that day.

        >>> pipeline.run_for_date(execution_date="2020-08-04")

    """

    def __init__(
        self,
        source: Source,
        feature_set: FeatureSet,
        sink: Sink,
        spark_client: Optional[SparkClient] = None,
    ):
        self.source = source
        self.feature_set = feature_set
        self.sink = sink
        self.spark_client = spark_client or SparkClient()

    @property
    def source(self) -> Source:
        """Source of the data, the entry point of the pipeline."""
        return self._source

    @source.setter
    def source(self, source: Source) -> None:
        if not isinstance(source, Source):
            raise ValueError("source must be a Source instance")
        self._source = source

    @property
    def feature_set(self) -> FeatureSet:
        """Feature set composed by features and context metadata."""
        return self._feature_set

    @feature_set.setter
    def feature_set(self, feature_set: FeatureSet) -> None:
        if not isinstance(feature_set, FeatureSet):
            raise ValueError("feature_set must be a FeatureSet instance")
        self._feature_set = feature_set

    @property
    def sink(self) -> Sink:
        """Sink used to write the output dataframe in the desired locations."""
        return self._sink

    @sink.setter
    def sink(self, sink: Sink) -> None:
        if not isinstance(sink, Sink):
            raise ValueError("sink must be a Sink instance")
        self._sink = sink

    @property
    def spark_client(self) -> SparkClient:
        """Client used to access Spark connection."""
        return self._spark_client

    @spark_client.setter
    def spark_client(self, spark_client: SparkClient) -> None:

        if not isinstance(spark_client, SparkClient):
            raise ValueError("spark_client must be a SparkClient instance")

        self._spark_client = spark_client

    def run(
        self,
        end_date: Optional[str] = None,
        partition_by: Optional[List[str]] = None,
        order_by: Optional[List[str]] = None,
        num_processors: Optional[int] = None,
        start_date: Optional[str] = None,
    ) -> None:
        """Runs the defined feature set pipeline.

        The pipeline consists in the following steps:
        - Constructs the input dataframe from the data source.
        - Construct the feature set dataframe using the defined Features.
        - Load the data to the configured sink locations.

        It's important to notice, however, that both parameters partition_by
        and num_processors are WIP, we intend to enhance their functionality
        soon. Use only if strictly necessary.

        """
        dataframe = self.source.construct(
            client=self.spark_client,
            start_date=self.feature_set.define_start_date(start_date),
            end_date=end_date,
        )

        if partition_by:
            order_by = order_by or partition_by
            dataframe = repartition_sort_df(
                dataframe, partition_by, order_by, num_processors
            )

        dataframe = self.feature_set.construct(
            dataframe=dataframe,
            client=self.spark_client,
            start_date=start_date,
            end_date=end_date,
            num_processors=num_processors,
        )

        self.sink.flush(
            dataframe=dataframe,
            feature_set=self.feature_set,
            spark_client=self.spark_client,
        )

        if not dataframe.isStreaming:
            self.sink.validate(
                dataframe=dataframe,
                feature_set=self.feature_set,
                spark_client=self.spark_client,
            )

    def run_for_date(
        self,
        execution_date: Optional[str] = None,
        partition_by: Optional[List[str]] = None,
        order_by: Optional[List[str]] = None,
        num_processors: Optional[int] = None,
    ) -> None:
        """Runs the defined feature set pipeline for a specific date.

        The pipeline consists in the following steps:

        - Constructs the input dataframe from the data source.
        - Construct the feature set dataframe using the defined Features.
        - Load the data to the configured sink locations.

        It's important to notice, however, that both parameters partition_by
        and num_processors are WIP, we intend to enhance their functionality
        soon. Use only if strictly necessary.
        """
        self.run(
            start_date=execution_date,
            end_date=execution_date,
            partition_by=partition_by,
            order_by=order_by,
            num_processors=num_processors,
        )



"""Holds the Historical Feature Store writer class."""

import os
from typing import Any, Optional

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import dayofmonth, month, year

from butterfree.clients import SparkClient
from butterfree.configs import environment
from butterfree.configs.db import AbstractWriteConfig, MetastoreConfig
from butterfree.constants import columns
from butterfree.constants.spark_constants import DEFAULT_NUM_PARTITIONS
from butterfree.dataframe_service import repartition_df
from butterfree.hooks import Hook
from butterfree.hooks.schema_compatibility import SparkTableSchemaCompatibilityHook
from butterfree.load.writers.delta_writer import DeltaWriter
from butterfree.load.writers.writer import Writer
from butterfree.transform import FeatureSet


class HistoricalFeatureStoreWriter(Writer):
    """Enable writing feature sets into the Historical Feature Store.

    Attributes:
        db_config: Datalake configuration for Spark, by default on AWS S3.
            For more information check module 'butterfree.db.configs'.
        database: database name to use in Spark metastore.
            By default FEATURE_STORE_HISTORICAL_DATABASE environment variable.
        num_partitions: value to use when applying repartition on the df before save.
        validation_threshold: lower and upper tolerance to using in count validation.
            The default value is defined in DEFAULT_VALIDATION_THRESHOLD property.
            For example: with a validation_threshold = 0.01 and a given calculated
            count on the dataframe equal to 100000 records, if the feature store
            return a count equal to 995000 an error will not be thrown.
            Use validation_threshold = 0 to not use tolerance in the validation.
        debug_mode: "dry run" mode, write the result to a temporary view.

    Example:
        Simple example regarding HistoricalFeatureStoreWriter class instantiation.
        We can instantiate this class without db configurations, so the class get the
        S3Config() where it provides default configurations about AWS S3 service.

    >>> spark_client = SparkClient()
    >>> writer = HistoricalFeatureStoreWriter()
    >>> writer.write(feature_set=feature_set,
       ...           dataframe=dataframe,
       ...           spark_client=spark_client)

        However, we can define the db configurations,
        like write mode, file format and S3 bucket,
        and provide them to HistoricalFeatureStoreWriter.

    >>> spark_client = SparkClient()
    >>> config = MetastoreConfig(path="my_s3_bucket_name",
        ...               mode="overwrite",
        ...               format_="parquet")
    >>> writer = HistoricalFeatureStoreWriter(db_config=config)
    >>> writer.write(feature_set=feature_set,
       ...           dataframe=dataframe,
       ...           spark_client=spark_client)

        For what settings you can use on S3Config and default settings,
        to read S3Config class.

        We can write with interval mode, where HistoricalFeatureStoreWrite
        will need to use Dynamic Partition Inserts,
        the behaviour of OVERWRITE keyword is controlled by
        spark.sql.sources.partitionOverwriteMode configuration property.
        The dynamic overwrite mode is enabled Spark will only delete the
        partitions for which it has data to be written to.
        All the other partitions remain intact.

    >>> spark_client = SparkClient()
    >>> writer = HistoricalFeatureStoreWriter(interval_mode=True)
    >>> writer.write(feature_set=feature_set,
       ...           dataframe=dataframe,
       ...           spark_client=spark_client)

        We can instantiate HistoricalFeatureStoreWriter class to validate the df
        to be written.

    >>> spark_client = SparkClient()
    >>> writer = HistoricalFeatureStoreWriter()
    >>> writer.validate(feature_set=feature_set,
       ...              dataframe=dataframe,
       ...              spark_client=spark_client)

        Both methods (write and validate) will need the Spark Client, Feature Set
        and DataFrame, to write or to validate, according to the Writer's arguments.

        P.S.: When writing, the HistoricalFeatureStoreWrite partitions the data to
        improve queries performance. The data is stored in partition folders in AWS S3
        based on time (per year, month and day).

        >>> spark_client = SparkClient()
        >>> writer = HistoricalFeatureStoreWriter()
        >>> writer.write(feature_set=feature_set,
        ...           dataframe=dataframe,
        ...           spark_client=spark_client
        ...           merge_on=["id", "timestamp"])

        This procedure will skip dataframe write and will activate Delta Merge.
        Use it when the table already exist.
    """

    PARTITION_BY = [
        columns.PARTITION_YEAR,
        columns.PARTITION_MONTH,
        columns.PARTITION_DAY,
    ]

    DEFAULT_VALIDATION_THRESHOLD = 0.01

    __name__ = "Historical Feature Store Writer"

    def __init__(
        self,
        db_config: Optional[AbstractWriteConfig] = None,
        database: Optional[str] = None,
        num_partitions: Optional[int] = None,
        validation_threshold: float = DEFAULT_VALIDATION_THRESHOLD,
        debug_mode: bool = False,
        interval_mode: bool = False,
        check_schema_hook: Optional[Hook] = None,
        row_count_validation: bool = True,
        merge_on: list = None,
    ):
        super(HistoricalFeatureStoreWriter, self).__init__(
            db_config or MetastoreConfig(),
            debug_mode,
            interval_mode,
            False,
            row_count_validation,
            merge_on,
        )
        self.database = database or environment.get_variable(
            "FEATURE_STORE_HISTORICAL_DATABASE"
        )
        self.num_partitions = num_partitions or DEFAULT_NUM_PARTITIONS
        self.validation_threshold = validation_threshold
        self.check_schema_hook = check_schema_hook

    def write(
        self,
        feature_set: FeatureSet,
        dataframe: DataFrame,
        spark_client: SparkClient,
    ) -> None:
        """Loads the data from a feature set into the Historical Feature Store.

        Args:
            feature_set: object processed with feature_set informations.
            dataframe: spark dataframe containing data from a feature set.
            spark_client: client for spark connections with external services.
            merge_on: when filled, the writing is an upsert in a Delta table.

        If the debug_mode is set to True, a temporary table with a name in the format:
        historical_feature_store__{feature_set.name} will be created instead of writing
        to the real historical feature store.

        """
        dataframe = self._create_partitions(dataframe)

        dataframe = self._apply_transformations(dataframe)

        if self.interval_mode:

            partition_overwrite_mode = spark_client.conn.conf.get(  # type: ignore
                "spark.sql.sources.partitionOverwriteMode"
            ).lower()

            if partition_overwrite_mode != "dynamic":
                raise RuntimeError(
                    "m=load_incremental_table, "
                    "spark.sql.sources.partitionOverwriteMode={}, "
                    "msg=partitionOverwriteMode have to "
                    "be configured to 'dynamic'".format(partition_overwrite_mode)
                )

        if self.debug_mode:
            spark_client.create_temporary_view(
                dataframe=dataframe,
                name=f"historical_feature_store__{feature_set.name}",
            )
            return

        s3_key = os.path.join("historical", feature_set.entity, feature_set.name)

        if self.merge_on:
            DeltaWriter.merge(
                client=spark_client,
                database=self.database,
                table=feature_set.name,
                merge_on=self.merge_on,
                source_df=dataframe,
            )
        else:
            spark_client.write_table(
                dataframe=dataframe,
                database=self.database,
                table_name=feature_set.name,
                partition_by=self.PARTITION_BY,
                **self.db_config.get_options(s3_key),
            )

    def _assert_validation_count(
        self, table_name: str, written_count: int, dataframe_count: int
    ) -> None:
        lower_bound = (1 - self.validation_threshold) * written_count
        upper_bound = (1 + self.validation_threshold) * written_count
        validation = lower_bound <= dataframe_count <= upper_bound
        assert validation, (
            "Data written to the Historical Feature Store and read back "
            f"from {table_name} has a different count than the feature set dataframe. "
            f"\nNumber of rows in {table_name}: {written_count}."
            f"\nNumber of rows in the dataframe: {dataframe_count}."
        )

    def validate(
        self, feature_set: FeatureSet, dataframe: DataFrame, spark_client: SparkClient
    ) -> None:
        """Calculate dataframe rows to validate data into Feature Store.

        Args:
            feature_set: object processed with feature_set informations.
            dataframe: spark dataframe containing data from a feature set.
            spark_client: client for spark connections with external services.

        Raises:
            AssertionError: if count of written data doesn't match count in current
                feature set dataframe.
        """
        table_name = (
            os.path.join("historical", feature_set.entity, feature_set.name)
            if self.interval_mode and not self.debug_mode
            else (
                f"{self.database}.{feature_set.name}"
                if not self.debug_mode
                else f"historical_feature_store__{feature_set.name}"
            )
        )

        written_count = (
            spark_client.read(
                self.db_config.format_,
                path=self.db_config.get_path_with_partitions(
                    table_name, self._create_partitions(dataframe)
                ),
            ).count()
            if self.interval_mode and not self.debug_mode
            else spark_client.read_table(table_name).count()
        )

        dataframe_count = dataframe.count()

        self._assert_validation_count(table_name, written_count, dataframe_count)

    def _create_partitions(self, dataframe: DataFrame) -> DataFrame:
        # create year partition column
        dataframe = dataframe.withColumn(
            columns.PARTITION_YEAR, year(dataframe[columns.TIMESTAMP_COLUMN])
        )
        # create month partition column
        dataframe = dataframe.withColumn(
            columns.PARTITION_MONTH, month(dataframe[columns.TIMESTAMP_COLUMN])
        )
        # create day partition column
        dataframe = dataframe.withColumn(
            columns.PARTITION_DAY, dayofmonth(dataframe[columns.TIMESTAMP_COLUMN])
        )
        return repartition_df(dataframe, self.PARTITION_BY, self.num_partitions)

    def check_schema(
        self,
        client: Any,
        dataframe: DataFrame,
        table_name: str,
        database: Optional[str] = None,
    ) -> DataFrame:
        """Instantiate the schema check hook to check schema between dataframe and database.

        Args:
            client: client for Spark or Cassandra connections with external services.
            dataframe: Spark dataframe containing data from a feature set.
            table_name: table name where the dataframe will be saved.
            database: database name where the dataframe will be saved.
        """
        if not self.check_schema_hook:
            self.check_schema_hook = SparkTableSchemaCompatibilityHook(
                client, table_name, database
            )

        return self.check_schema_hook.run(dataframe)



"""Spark table schema compatibility Hook definition."""

from typing import Optional

from pyspark.sql import DataFrame

from butterfree.clients import SparkClient
from butterfree.hooks.hook import Hook


class SparkTableSchemaCompatibilityHook(Hook):
    """Hook to verify the schema compatibility with a Spark's table.

    Verifies if all columns presented on the dataframe exists and are the same
    type on the target Spark's table.

    Attributes:
        spark_client: client to connect to Spark's metastore.
        table: table name.
        database: database name.
    """

    def __init__(
        self, spark_client: SparkClient, table: str, database: Optional[str] = None
    ):
        self.spark_client = spark_client
        self.table_expression = (f"`{database}`." if database else "") + f"`{table}`"

    def run(self, dataframe: DataFrame) -> DataFrame:
        """Check the schema compatibility from a given Dataframe.

        This method does not change anything on the Dataframe.

        Args:
            dataframe: dataframe to verify schema compatibility.

        Returns:
            unchanged dataframe.

        Raises:
            ValueError if the schemas are incompatible.
        """
        table_schema = self.spark_client.conn.table(self.table_expression).schema
        if not all([column in table_schema for column in dataframe.schema]):
            raise ValueError(
                "The dataframe has a schema incompatible with the defined table.\n"
                f"Dataframe schema = {dataframe.schema}"
                f"Target table schema = {table_schema}"
            )
        return dataframe



"""Holds function for defining window in DataFrames."""

from typing import Any, List, Optional, Union

from pyspark import sql
from pyspark.sql import Column, WindowSpec, functions

from butterfree.constants.columns import TIMESTAMP_COLUMN
from butterfree.constants.window_definitions import ALLOWED_WINDOWS


class FrameBoundaries:
    """Utility functions for defining the frame boundaries.

    Args:
        mode: available modes to be used in time aggregations.
        window_definition: time ranges to be used in the windows,
        it can be second(s), minute(s), hour(s), day(s), week(s) and year(s),
    """

    def __init__(self, mode: Optional[str], window_definition: str):
        self.mode = mode
        self.window_definition = window_definition

    @property
    def window_size(self) -> int:
        """Returns window size."""
        if int(self.window_definition.split()[0]) <= 0:
            raise KeyError(f"{self.window_definition} have negative element.")
        return int(self.window_definition.split()[0])

    @property
    def window_unit(self) -> str:
        """Returns window unit."""
        unit = self.window_definition.split()[1]
        if unit not in ALLOWED_WINDOWS and self.mode != "row_windows":
            raise ValueError("Not allowed")

        return unit

    def get(self, window: WindowSpec) -> Any:
        """Returns window with or without the frame boundaries."""
        if self.mode is None:
            return window
        if self.mode == "row_windows":
            span = self.window_size - 1
            return window.rowsBetween(-span, 0)
        if self.mode == "fixed_windows":
            span = ALLOWED_WINDOWS[self.window_unit] * self.window_size
            return window.rangeBetween(-span, 0)


class Window:
    """Utility functions for defining a window specification.

    Args:
        partition_by: he partitioning defined.
        order_by: the ordering defined.
        mode: available modes to be used in time aggregations.
        window_definition: time ranges to be used in the windows, it can be second(s),
            minute(s), hour(s), day(s), week(s) and year(s),

    Use the static methods in :class:`Window` to create a :class:`WindowSpec`.
    """

    DEFAULT_SLIDE_DURATION: str = "1 day"

    def __init__(
        self,
        window_definition: str,
        partition_by: Optional[Union[Column, str, List[str]]] = None,
        order_by: Optional[Union[Column, str]] = None,
        mode: Optional[str] = None,
        slide: Optional[str] = None,
    ):
        self.partition_by = partition_by
        self.order_by = order_by or TIMESTAMP_COLUMN
        self.frame_boundaries = FrameBoundaries(mode, window_definition)
        self.slide = slide or self.DEFAULT_SLIDE_DURATION

    def get_name(self) -> str:
        """Return window suffix name based on passed criteria."""
        return "_".join(
            [
                "over",
                f"{self.frame_boundaries.window_size}",
                f"{self.frame_boundaries.window_unit}",
                self.frame_boundaries.mode,
            ]
        )

    def get(self) -> Any:
        """Defines a common window to be used both in time and rows windows."""
        if self.frame_boundaries.mode == "rolling_windows":
            return functions.window(
                TIMESTAMP_COLUMN,
                self.frame_boundaries.window_definition,
                slideDuration=self.slide,
            )
        elif self.order_by == TIMESTAMP_COLUMN:
            w = sql.Window.partitionBy(self.partition_by).orderBy(  # type: ignore
                functions.col(TIMESTAMP_COLUMN).cast("long")
            )
        else:
            w = sql.Window.partitionBy(self.partition_by).orderBy(  # type: ignore
                self.order_by
            )
        return self.frame_boundaries.get(w)



"""Transform Abstract Class."""

from abc import ABC, abstractmethod
from typing import Any, List

from pyspark.sql import DataFrame


class TransformComponent(ABC):
    """Defines an abstract class for Transform entities.

    Attributes:
        parent: parent transform component.
    """

    def __init__(self) -> None:
        self._parent = None

    @property
    def parent(self) -> Any:
        """Parent transform component."""
        return self._parent

    @parent.setter
    def parent(self, parent: None) -> None:
        self._parent = parent

    @property
    @abstractmethod
    def output_columns(self) -> List[str]:
        """Columns generated by the transformation."""

    @abstractmethod
    def transform(self, dataframe: DataFrame) -> DataFrame:
        """Performs a transformation to the feature pipeline.

        Args:
            dataframe: input dataframe.

        Returns:
            Transformed dataframe.
        """



"""Json conversion for writers."""

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import struct, to_json


def json_transform(dataframe: DataFrame) -> DataFrame:
    """Filters DataFrame's rows using the given condition and value.

    Args:
        dataframe: Spark DataFrame.

    Returns:
        Converted dataframe.
    """
    return dataframe.select(
        to_json(
            struct([dataframe[column] for column in dataframe.columns])  # type: ignore
        ).alias("value")
    )



"""Holds functions for managing the running environment."""

import os
from typing import Optional

specification = {
    "ENVIRONMENT": "dev",
    "CASSANDRA_HOST": "test",
    "CASSANDRA_KEYSPACE": "test",
    "CASSANDRA_USERNAME": "test",
    "CASSANDRA_PASSWORD": "test",
    "FEATURE_STORE_S3_BUCKET": "test",
    "FEATURE_STORE_HISTORICAL_DATABASE": "test",
    "KAFKA_CONSUMER_CONNECTION_STRING": "test_host:1234,test_host2:1234",
    "STREAM_CHECKPOINT_PATH": None,
    "CASSANDRA_READ_CONSISTENCY_LEVEL": None,
    "CASSANDRA_WRITE_CONSISTENCY_LEVEL": None,
    "CASSANDRA_LOCAL_DC": None,
}


class UnspecifiedVariableError(RuntimeError):
    """Environment variables not set error.

    Attributes:
        variable_name: environment variable name.

    """

    def __init__(self, variable_name: str):
        super().__init__(
            f'Variable "{variable_name}" is not listed in the environment'
            " specification\nUpdate the environment module"
            f' to include "{variable_name}"'
        )


def get_variable(
    variable_name: str, default_value: Optional[str] = None
) -> Optional[str]:
    """Gets an environment variable.

    The variable comes from it's explicitly declared value in the running
    environment or from the default value declared in specification or from the
    default_value.

    Args:
        variable_name: environment variable name.
        default_value: default value to use in case no value is set in the
            environment nor in the environment.yaml specification file.

    Returns:
        The variable's string value

    """
    try:
        spec_default = specification[variable_name]
    except KeyError:
        raise UnspecifiedVariableError(variable_name)
    return os.getenv(variable_name) or spec_default or default_value



"""The Reader Component of a Source."""

from butterfree.extract.readers.file_reader import FileReader
from butterfree.extract.readers.kafka_reader import KafkaReader
from butterfree.extract.readers.table_reader import TableReader

__all__ = ["FileReader", "KafkaReader", "TableReader"]



"""TimestampFeature entity."""

from typing import Optional

from pyspark.sql import DataFrame
from pyspark.sql.functions import to_timestamp

from butterfree.constants import DataType
from butterfree.constants.columns import TIMESTAMP_COLUMN
from butterfree.transform.features import Feature
from butterfree.transform.transformations import TransformComponent


class TimestampFeature(Feature):
    """Defines a TimestampFeature.

    A FeatureSet must contain one TimestampFeature, which will be used as a time
    tag for the state of all features. By containing a timestamp feature, users
    may time travel over their features. The Feature Set may validate that the
    set of keys and timestamp are unique for a feature set.

    By defining a TimestampColumn, the feature set will always contain a data
    column called "timestamp" of TimestampType (spark dtype).

    Attributes:
        from_column: original column to build a "timestamp" feature column.
            Used when there is transformation or the transformation has no
            reference about the column to use for.
            If from_column is None, the FeatureSet will assume the input
            dataframe already has a data column called "timestamp".
        transformation: transformation that will be applied to create the
            "timestamp". Type casting will already happen when no transformation
            is given. But a timestamp can be derived from multiple columns, like
            year, month and day, for example. The transformation must always
            handle naming and typing.
        from_ms: true if timestamp column presents milliseconds time unit. A
        conversion is then performed.
        mask: specified timestamp format by the user.

    """

    def __init__(
        self,
        dtype: Optional[DataType] = DataType.TIMESTAMP,
        from_column: Optional[str] = None,
        transformation: Optional[TransformComponent] = None,
        from_ms: bool = False,
        mask: Optional[str] = None,
    ) -> None:
        description = "Time tag for the state of all features."
        super(TimestampFeature, self).__init__(
            name=TIMESTAMP_COLUMN,
            description=description,
            from_column=from_column,
            dtype=dtype,
            transformation=transformation,
        )
        self.from_ms = from_ms
        self.mask = mask

    def transform(self, dataframe: DataFrame) -> DataFrame:
        """Performs a transformation to the feature pipeline.

        Args:
            dataframe: input dataframe for the transformation.

        Returns:
            Transformed dataframe.
        """
        column_name = self.from_column if self.from_column else self.name

        ts_column = dataframe[column_name]
        if self.from_ms:
            ts_column = ts_column / 1000

        dataframe = dataframe.withColumn(
            column_name, to_timestamp(ts_column, self.mask)  # type: ignore
        )

        return super().transform(dataframe)



"""Utils for date range generation."""

from datetime import datetime
from typing import Optional, Union

from pyspark.sql import DataFrame, functions

from butterfree.clients import SparkClient
from butterfree.constants import DataType
from butterfree.constants.columns import TIMESTAMP_COLUMN


def get_date_range(
    client: SparkClient,
    start_date: Union[str, datetime],
    end_date: Union[str, datetime],
    step: Optional[int] = None,
) -> DataFrame:
    """Create a date range dataframe.

    The dataframe returning from this method will containing a single column
    TIMESTAMP_COLUMN, of timestamp type, with dates between start and end.

    Args:
        client: a spark client.
        start_date: range beginning value (inclusive).
        end_date: range last value (exclusive)
        step: optional step, in seconds.

    Returns:
        A single column date range spark dataframe.
    """
    day_in_seconds = 60 * 60 * 24
    step = step or day_in_seconds
    start_date = (
        start_date if isinstance(start_date, str) else start_date.strftime("%Y-%m-%d")
    )
    end_date = end_date if isinstance(end_date, str) else end_date.strftime("%Y-%m-%d")
    date_df = client.conn.createDataFrame(
        [(start_date, end_date)], ("start_date", "end_date")
    ).select(
        [
            functions.col(c).cast(DataType.TIMESTAMP.spark).cast(DataType.BIGINT.spark)
            for c in ("start_date", "end_date")
        ]
    )
    start_date, end_date = date_df.first()  # type: ignore
    return client.conn.range(
        start_date, end_date + day_in_seconds, step  # type: ignore
    ).select(functions.col("id").cast(DataType.TIMESTAMP.spark).alias(TIMESTAMP_COLUMN))



"""Holds configurations to read and write with Spark to Cassandra DB."""

from typing import Any, Dict, List, Optional

from butterfree.configs import environment
from butterfree.configs.db import AbstractWriteConfig


class CassandraConfig(AbstractWriteConfig):
    """Configuration for Spark to connect on Cassandra DB.

    References can be found
    [here](https://docs.databricks.com/data/data-sources/cassandra.html).

    Attributes:
        username: username to use in connection.
        password: password to use in connection.
        host: host to use in connection.
        keyspace:  Cassandra DB keyspace to write data.
        mode: write mode for Spark.
        format_: write format for Spark.
        stream_processing_time: processing time interval for streaming jobs.
        stream_output_mode: specify the mode from writing streaming data.
        stream_checkpoint_path: path on S3 to save checkpoints for the stream job.
        read_consistency_level: read consistency level used in connection.
        write_consistency_level: write consistency level used in connection.

    More information about processing_time, output_mode and checkpoint_path
    can be found in Spark documentation:
    [here](https://spark.apache.org/docs/latest/structured-streaming-programming-guide.html)

    """

    def __init__(
        self,
        username: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        keyspace: Optional[str] = None,
        mode: Optional[str] = None,
        format_: Optional[str] = None,
        stream_processing_time: Optional[str] = None,
        stream_output_mode: Optional[str] = None,
        stream_checkpoint_path: Optional[str] = None,
        read_consistency_level: Optional[str] = None,
        write_consistency_level: Optional[str] = None,
        local_dc: Optional[str] = None,
    ):
        self.username = username
        self.password = password
        self.host = host
        self.keyspace = keyspace
        self.mode = mode
        self.format_ = format_
        self.stream_processing_time = stream_processing_time
        self.stream_output_mode = stream_output_mode
        self.stream_checkpoint_path = stream_checkpoint_path
        self.read_consistency_level = read_consistency_level
        self.write_consistency_level = write_consistency_level
        self.local_dc = local_dc

    @property
    def database(self) -> str:
        """Database name."""
        return "cassandra"

    @property
    def username(self) -> Optional[str]:
        """Username used in connection to Cassandra DB."""
        return self.__username

    @username.setter
    def username(self, value: str) -> None:
        input_value = value or environment.get_variable("CASSANDRA_USERNAME")
        if input_value is None:
            raise ValueError("Config 'username' cannot be empty.")
        self.__username = input_value

    @property
    def password(self) -> Optional[str]:
        """Password used in connection to Cassandra DB."""
        return self.__password

    @password.setter
    def password(self, value: str) -> None:
        input_value = value or environment.get_variable("CASSANDRA_PASSWORD")
        if input_value is None:
            raise ValueError("Config 'password' cannot be empty.")
        self.__password = input_value

    @property
    def host(self) -> Optional[str]:
        """Host used in connection to Cassandra DB."""
        return self.__host

    @host.setter
    def host(self, value: str) -> None:
        input_value = value or environment.get_variable("CASSANDRA_HOST")
        if input_value is None:
            raise ValueError("Config 'host' cannot be empty.")
        self.__host = input_value

    @property
    def keyspace(self) -> Optional[str]:
        """Cassandra DB keyspace to write data."""
        return self.__keyspace

    @keyspace.setter
    def keyspace(self, value: str) -> None:
        input_value = value or environment.get_variable("CASSANDRA_KEYSPACE")
        if not input_value:
            raise ValueError("Config 'keyspace' cannot be empty.")
        self.__keyspace = input_value

    @property
    def format_(self) -> Optional[str]:
        """Write format for Spark."""
        return self.__format

    @format_.setter
    def format_(self, value: str) -> None:
        self.__format = value or "org.apache.spark.sql.cassandra"

    @property
    def mode(self) -> Optional[str]:
        """Write mode for Spark."""
        return self.__mode

    @mode.setter
    def mode(self, value: str) -> None:
        self.__mode = value or "append"

    @property
    def stream_processing_time(self) -> Optional[str]:
        """Processing time interval for streaming jobs."""
        return self.__stream_processing_time

    @stream_processing_time.setter
    def stream_processing_time(self, value: str) -> None:
        self.__stream_processing_time = value or "0 seconds"

    @property
    def stream_output_mode(self) -> Optional[str]:
        """Specify the mode from writing streaming data."""
        return self.__stream_output_mode

    @stream_output_mode.setter
    def stream_output_mode(self, value: str) -> None:
        self.__stream_output_mode = value or "update"

    @property
    def stream_checkpoint_path(self) -> Optional[str]:
        """Path on S3 to save checkpoints for the stream job."""
        return self.__stream_checkpoint_path

    @stream_checkpoint_path.setter
    def stream_checkpoint_path(self, value: str) -> None:
        self.__stream_checkpoint_path = value or environment.get_variable(
            "STREAM_CHECKPOINT_PATH"
        )

    @property
    def read_consistency_level(self) -> Optional[str]:
        """Read consistency level for Cassandra."""
        return self.__read_consistency_level

    @read_consistency_level.setter
    def read_consistency_level(self, value: str) -> None:
        self.__read_consistency_level = value or environment.get_variable(
            "CASSANDRA_READ_CONSISTENCY_LEVEL", "LOCAL_ONE"
        )

    @property
    def write_consistency_level(self) -> Optional[str]:
        """Write consistency level for Cassandra."""
        return self.__write_consistency_level

    @write_consistency_level.setter
    def write_consistency_level(self, value: str) -> None:
        self.__write_consistency_level = value or environment.get_variable(
            "CASSANDRA_WRITE_CONSISTENCY_LEVEL", "LOCAL_QUORUM"
        )

    @property
    def local_dc(self) -> Optional[str]:
        """Local DC for Cassandra connection."""
        return self.__local_dc

    @local_dc.setter
    def local_dc(self, value: str) -> None:
        self.__local_dc = value or environment.get_variable("CASSANDRA_LOCAL_DC")

    def get_options(self, table: str) -> Dict[Optional[str], Optional[str]]:
        """Get options for connect to Cassandra DB.

        Options will be a dictionary with the write and read configuration for
        spark to cassandra.

        Args:
            table: table name (keyspace) into Cassandra DB.

        Returns:
            Configuration to connect to Cassandra DB.

        """
        return {
            "table": table,
            "keyspace": self.keyspace,
            "spark.cassandra.auth.username": self.username,
            "spark.cassandra.auth.password": self.password,
            "spark.cassandra.connection.host": self.host,
            "spark.cassandra.connection.localDC": self.local_dc,
            "spark.cassandra.input.consistency.level": self.read_consistency_level,
            "spark.cassandra.output.consistency.level": self.write_consistency_level,
        }

    def translate(self, schema: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """Get feature set schema to be translated.

        The output will be a list of dictionaries regarding cassandra
        database schema.

        Args:
            schema: feature set schema in spark.

        Returns:
            Cassandra schema.

        """
        cassandra_mapping = {
            "timestamptype": "timestamp",
            "timestampntztype": "timestamp",
            "binarytype": "boolean",
            "booleantype": "boolean",
            "datetype": "timestamp",
            "decimaltype": "decimal",
            "doubletype": "double",
            "floattype": "float",
            "integertype": "int",
            "longtype": "bigint",
            "stringtype": "text",
            "arraytype(longtype, true)": "frozen<list<bigint>>",
            "arraytype(stringtype, true)": "frozen<list<text>>",
            "arraytype(floattype, true)": "frozen<list<float>>",
        }
        cassandra_schema = []
        for features in schema:
            cassandra_schema.append(
                {
                    "column_name": features["column_name"],
                    "type": cassandra_mapping[
                        str(features["type"]).replace("()", "").lower()
                    ],
                    "primary_key": features["primary_key"],
                }
            )

        return cassandra_schema



"""Definition of hookable component."""

from __future__ import annotations

from typing import List

from pyspark.sql import DataFrame

from butterfree.hooks.hook import Hook


class HookableComponent:
    """Defines a component with the ability to hold pre and post hook functions.

    All main module of Butterfree have a common object that enables their integration:
    dataframes. Spark's dataframe is the glue that enables the transmission of data
    between the main modules. Hooks have a simple interface, they are functions that
    accepts a dataframe and outputs a dataframe. These Hooks can be triggered before or
    after the main execution of a component.

    Components from Butterfree that inherit HookableComponent entity, are components
    that can define a series of steps to occur before or after the execution of their
    main functionality.

    Attributes:
        pre_hooks: function steps to trigger before component main functionality.
        post_hooks: function steps to trigger after component main functionality.
        enable_pre_hooks: property to indicate if the component can define pre_hooks.
        enable_post_hooks: property to indicate if the component can define post_hooks.
    """

    def __init__(self) -> None:
        self.pre_hooks = []
        self.post_hooks = []
        self.enable_pre_hooks = True
        self.enable_post_hooks = True

    @property
    def pre_hooks(self) -> List[Hook]:
        """Function steps to trigger before component main functionality."""
        return self.__pre_hook

    @pre_hooks.setter
    def pre_hooks(self, value: List[Hook]) -> None:
        if not isinstance(value, list):
            raise ValueError("pre_hooks should be a list of Hooks.")
        if not all(isinstance(item, Hook) for item in value):
            raise ValueError(
                "All items on pre_hooks list should be an instance of Hook."
            )
        self.__pre_hook = value

    @property
    def post_hooks(self) -> List[Hook]:
        """Function steps to trigger after component main functionality."""
        return self.__post_hook

    @post_hooks.setter
    def post_hooks(self, value: List[Hook]) -> None:
        if not isinstance(value, list):
            raise ValueError("post_hooks should be a list of Hooks.")
        if not all(isinstance(item, Hook) for item in value):
            raise ValueError(
                "All items on post_hooks list should be an instance of Hook."
            )
        self.__post_hook = value

    @property
    def enable_pre_hooks(self) -> bool:
        """Property to indicate if the component can define pre_hooks."""
        return self.__enable_pre_hooks

    @enable_pre_hooks.setter
    def enable_pre_hooks(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise ValueError("enable_pre_hooks accepts only boolean values.")
        self.__enable_pre_hooks = value

    @property
    def enable_post_hooks(self) -> bool:
        """Property to indicate if the component can define post_hooks."""
        return self.__enable_post_hooks

    @enable_post_hooks.setter
    def enable_post_hooks(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise ValueError("enable_post_hooks accepts only boolean values.")
        self.__enable_post_hooks = value

    def add_pre_hook(self, *hooks: Hook) -> HookableComponent:
        """Add a pre-hook steps to the component.

        Args:
            hooks: Hook steps to add to pre_hook list.

        Returns:
            Component with the Hook inserted in pre_hook list.

        Raises:
            ValueError: if the component does not accept pre-hooks.
        """
        if not self.enable_pre_hooks:
            raise ValueError("This component does not enable adding pre-hooks")
        self.pre_hooks += list(hooks)
        return self

    def add_post_hook(self, *hooks: Hook) -> HookableComponent:
        """Add a post-hook steps to the component.

        Args:
            hooks: Hook steps to add to post_hook list.

        Returns:
            Component with the Hook inserted in post_hook list.

        Raises:
            ValueError: if the component does not accept post-hooks.
        """
        if not self.enable_post_hooks:
            raise ValueError("This component does not enable adding post-hooks")
        self.post_hooks += list(hooks)
        return self

    def run_pre_hooks(self, dataframe: DataFrame) -> DataFrame:
        """Run all defined pre-hook steps from a given dataframe.

        Args:
            dataframe: data to input in the defined pre-hook steps.

        Returns:
            dataframe after passing for all defined pre-hooks.
        """
        for hook in self.pre_hooks:
            dataframe = hook.run(dataframe)
        return dataframe

    def run_post_hooks(self, dataframe: DataFrame) -> DataFrame:
        """Run all defined post-hook steps from a given dataframe.

        Args:
            dataframe: data to input in the defined post-hook steps.

        Returns:
            dataframe after passing for all defined post-hooks.
        """
        for hook in self.post_hooks:
            dataframe = hook.run(dataframe)
        return dataframe
